#!/usr/bin/env ruby
$:.push(  File.dirname( __FILE__ ) )
require 'webrick'
require 'logger'
require "timeout"




include WEBrick

module Neurogami

  # Simple WEBrick wrapper to allow dynamic loading and execution of code via
  # HTTP.
  # 
  # Catapult starts up a WEBrick instance, and transforms URL strings into
  # code invocations.  The first segment of a URL path is assumed to be both a
  # class and a file name.  The rest of the URL is passed on as parameters.
  # For example:
  #  http://example.com:9090/someApp/foo/bar?x=1&y=2
  # tells Catapult to create or reuse an instance of SomeApp and send the
  # message +run+, passing along +foo/bar?x=1&y=2+
  #
  # Some things to note:
  # * Catapult looks to require files from the default Ruby lib path, as well
  #   as in any directories defined in the environment variable CATAPULT_PATH
  # * The class/file name passed in the URL may start with either an upper or
  #   lower case letter, but Catapult will always convert that first letter to
  #   upper case.  Files and classes must therefore have the same exact name; the
  #   use of lower case in the URL is for aesthetic reasons.
  # * Objects called by Catapult must respond to three messages: 
  #    run( some_string ) -> response string
  #    content_type() -> the HTTP content type string
  #    shutdown() -> nil
  # * Catapult caches objects to avoid recreating an instance on every call.  
  #   This means you should be able to initialize an object just once, but it
  #   also means that the code called by Catapult cannot rely on getting fresh
  #   values passed in via 'new'.  However, the code called by Catapult can in
  #   turn wrap some other code that gets re-instantiated as often as you like.
  # * Once Catapult is running, you should be able to just drop new code into
  #   any directory in the Catapult load path and have it Just Work, without
  #   restarting Catapult.
  #
  # All objects known to Catapult have access to the ObjectCache as well as each 
  # other.  This allows Catapult to bootstrap its features.  For example, an
  # early version had Catapult mapping a specie URL path to  a proc for
  # clearing objects from the ObjectCache.  It worked, but the implementation
  # involved a fixed URL, which seemed a bad idea.  In considering a new
  # approach, it seemed that, as nice a feature as this is, it bordered on
  # feature creep.  Since the URL should be flexible, it could be defined by
  # creating a new object for Catapult to load.  As these objects have access
  # to both the Catapult and ObjectCache classes, all the needed code could
  # live in this dynamically loaded object.
  class Catapult

    def self.allow_from? ip
      CataLog.debug( "allow_from has allow_from_list = #{@@allow_from_list.inspect} and ip #{ip.to_s.strip}" )
      return true unless @@allow_from_list
      return true unless @@allow_from_list.to_s.size > 0
      CataLog.debug( "allow_from to see if #{@@allow_from_list.inspect} contains #{ip.to_s.strip}" )
      @@allow_from_list.each do |req_ip|
        return true if req_ip.to_s.strip == ip.to_s.strip
      end
      false
    end
    
    def initialize( options={} )
      set_require_path
      @port = options[ :port ] || 9090
      @shutdown_wait = options[ :shutdown_wait ] || 10.0
      @document_root = options[ :document_root ] || Dir.pwd
      @quit_path = options[ :quit_path ] || '/quit'
      @@allow_from_list = options[ :allow_from ] || nil
      catabug( "Catapult initialized with port = #@port, and document_root = #@document_root" )
    end

    def set_require_path
      return unless  ENV[ 'CATAPULT_PATH' ]
      /mswin/ =~ RUBY_PLATFORM ? delim  = ';' :  delim  = ':' 
      STDERR.puts( "Have ENV[ 'CATAPULT_PATH' ] = #{ENV[ 'CATAPULT_PATH' ]}")
      ENV[ 'CATAPULT_PATH' ].split( delim  ).each { |path| $:.push( path ) }
    end
    
    def run
      begin
        s = HTTPServer.new(
          :Port            => @port,
          :DocumentRoot    => @document_root )
        s.mount( "/",  RequestHandler  )
        s.mount( "/favicon.ico",   nil ) # Pet peeve.
        s.mount_proc( @quit_path ) do |req, resp| 
          begin
            ObjectCache.shutdown(  @shutdown_wait )
          rescue Exception
            CataLog.error "Error shutting down cached objects: #{$!}"          
          end
          s.shutdown;  exit;  
        end  
        
        trap( 1 ){ s.shutdown }
        s.start
      rescue Exception
        CataLog.error "Error in Catapult: #{$!}"
        s.shutdown
      end    
    end

  end

    # Wrapper for a gloabal logger.
    # If Catapult ends up being run as a Windows service or something similar, then
    # it may make sense to replace file logging with something more approriate
    # to the processing environment, such as an event log.  But you likely will not
    # want scads of debug messages going there, so be sure the log level or
    # whatever is set accordingly.
  class CataLog
    @@catalog = Logger.new( "cata.log"  )
    @@catalog.level = Logger::DEBUG

    def self.method_missing( meth , *args )
      @@catalog.send( meth , *args )
    end
  end

  def catabug( str )
    CataLog.debug( str )
  end

    # Rather than instantiate a new object for each call, we can just reuse
    # objects that have been called at least once.  Of course, this means that
    # any code meant to be invoked by Catapult cannot rely on per-call
    # initialization via 'new'.
    # The actual class loaded by Catapult can simply wrap other objects, so
    # there should not be any issues if the main business objects really do need
    # to instaniated from scratch on each invocation.
  class ObjectCache
    
    @@cache = {}
    
      # Returns an instance of the class named by cls_name.  It first looks in
      # the cache, and if not found, tries to make a new one, then stashes this
      # new one in the cache before returning it.
    def self.get( classs_name )
      return @@cache[ classs_name.strip  ] if  @@cache[ classs_name.strip  ]
      obj = ObjectCache.new_from_name( classs_name )
      @@cache[ classs_name.strip  ] = obj
      STDERR.puts( "cache now = #{names.inspect}")
      obj
    end
    
      # Adds an object to the cache.  You can also pass an explicit key (assumed
      # to be the class name, or you will not be able to fetch it).  If you do
      # not pass in this name, the class type is used asa  default.
    def puts( klass, class_name = nil )

      class_name ||= klass.type.to_s
      STDERR.puts( "put #{class_name}")    
      @@cache[ class_name ] = klass
    end

      # Instantiates a new instance of the class named by class_name.
    def self.new_from_name( class_name )
      cname = String.new( class_name.untaint )
      obj = nil
      begin
        catabug( "Try to load '#{cname}'")
        load cname + ".rb"
        obj = Object.const_get( cname ).new
        
      rescue Exception
        catabug( "Error: Cannot create object #{cname}: #{$!}
                  Tried loading from \n#{$:.join( "\n")}" ) unless obj
        
        raise "Error: Cannot create object #{cname}: #{$!}
               Tried loading from \n#{$:.join( "\n")}" unless obj
      end
      obj
    end
      # Loops over all objects in the cache and invokes +shutdown+ on each one.
      # Uses the +Timeout+ lib to control so that any shutdown taking more than
      # seconds_to_wait rasies and exception; that exception is essentially
      # ignored here.  
    def self.shutdown( seconds_to_wait = 5.0 )
      @@cache.each do |k,v|
        begin
          timeout( seconds_to_wait.to_f ) do
            v.send( 'shutdown' )
          end
        rescue Exception
          CataLog.warn( "Error calling shutdown for #{k}.")        
        end
      end
    end

    def self.clear( class_name = nil )

      if class_name
        CataLog.debug( "Calling shutdown for #{class_name}.")
        begin
          @@cache[ class_name.strip ].send( 'shutdown' )
        rescue Exception; end
        @@cache.delete( class_name.strip )
        
      else
        @@cache.each_key do |class_name|
        CataLog.debug( "Calling shutdown for #{class_name}.")              
        begin
          @@cache[ class_name.strip ].send( 'shutdown' )        
        rescue Exception; end
        sleep( 10 )
        @@cache = {}
        end
      end
      
    end

    def self.names
      @@cache.keys
    end
    
  end
end
  # This is the catch-all request handler class.  It is invoked on either a GET or a
  # POST, extracts a class name from the URL's unparsed_uri, and calls into that
  # object to do the actual work, relaying back the results.
class RequestHandler < HTTPServlet::AbstractServlet

    # Delegates to process_request
  def do_GET(req, res)
    process_request( req, res , :get)
  end
  
    # Delegates to process_request  
  def do_POST(req, res)
    process_request( req, res, :post )
  end

  def not_allowed( request, response )

    response.body = <<-DEFAULT
    <html>
     <head>
      <title>Catapult error</title>
     </head>
     <body>
      <h1>Catapult Error</h1>
      <p>Forbidden!</p>
     </body>
    </html> 
    DEFAULT
    response[ 'content-type'] = "text/html"
  end

  def exception( request, response, execption )

    response.body = <<-DEFAULT
    <html>
     <head>
      <title>Catapult error</title>
     </head>
     <body>
      <h1>Catapult Error</h1>
      <p>Exception:</p>
      <pre>#{execption.to_s}</pre>
     </body>
    </html> 
    DEFAULT
    response[ 'content-type'] = "text/html"
  end   

    # Creates the actual business logic object by extracting a class name from
    # the unparsed_uri value:
    # /class_name/then/come/paramters?and=query&string=stuff
  def process_request( request, response, rq_type = :get )
    Neurogami::CataLog.debug( "process_request request.peeraddr  = #{request.peeraddr.inspect }" ) 
    Neurogami::CataLog.debug( "process_request request  = #{request.inspect }" ) 
    Neurogami::CataLog.debug( "process_request request.unparsed_uri  = #{request. unparsed_uri}" ) 
    STDERR.puts( "process_request request.unparsed_uri  = #{request. unparsed_uri}" ) 
    
    #  example: ["AF_INET", 1833, "james2000", "127.0.0.1"]
    # Can we use this to control access?  
    return not_allowed( request, response ) unless  Catapult.allow_from? request.peeraddr[2].strip 
    return default_response( request, response ) unless request.unparsed_uri.to_s.size > 1 
   
    begin
      catabug( "Catapult#process_request has unparsed_uri #{request.unparsed_uri}" )
      path_parts = request.unparsed_uri.split( '/' ).reject { |x| x.to_s.strip.size == 0 }
      cls_name = path_parts.shift.to_s.strip
      cls_name[0] = cls_name[0].chr.upcase
      STDERR.puts( "Get '#{cls_name}'")
      obj = ObjectCache.get( cls_name )
      

      rq_type == :get ? req_qs = request.query.to_qs  : req_qs =  '/' + CGI.escape( request.body )
      response.body = obj.run( path_parts.join( '/' ) + req_qs )
      
      response[ 'content-type'] = obj.content_type
    rescue Exception
       exception( request, response, $! )      
    end
  end


  def default_response( request, response )
    response.body = <<-DEFAULT
    <html>
     <head>
      <title>Catapult default response</title>
     </head>
     <body>
      <h1>Catapult</h1>
      <p>It appears you did not specify any object to invoke.</p>
     </body>
    </html> 
    DEFAULT
    response[ 'content-type'] = "text/html"
  end
end


class Hash  # :nodoc
  def to_qs
    return '' unless self.keys.size > 0
    qs = self.keys.map { |k| "#{k}=#{self[k]}" }.join( '&') + ''
    qs.strip.size > 0     ? "?#{qs}" : ""
  end
end


if __FILE__ == $0 
  port =  ARGV[0] || 8989
  Catapult.new( :port=>port.to_i, 
                :shutdown_wait=>10 ).run()
end

#-- -----------------------------------------------------------------
# Copyright 2005 James Britt <james_b@neurogami.com> and Neurogami, LLC
# 
#  LICENSE
# 
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  To read the license please visit http://www.gnu.org/copyleft/gpl.html
#-- -----------------------------------------------------------------
#++
